/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <VirtualRobot/Robot.h>


namespace armarx::control::common::control_law
{

    struct RobotStatus
    {
        std::vector<float> jointPosition;
        std::vector<float> jointVelocity;
        std::vector<float> jointTorque;
        double deltaT = 0;
    };

    struct VirtualRobotStatus
    {
        Eigen::Matrix4f globalPose;
        std::vector<float> jointValuesFullRobot;
    };

    struct ControlTarget
    {
        std::vector<float> targets;
    };

    inline Eigen::MatrixXf
    computeJointLimitWeightMatrix(const Eigen::Ref<const Eigen::VectorXf>& jointAngles,
                                  const Eigen::Ref<const Eigen::VectorXf>& jointLimitsLow,
                                  const Eigen::Ref<const Eigen::VectorXf>& jointLimitsHigh,
                                  Eigen::VectorXf& jointAngleLimitGradient)
    {
        // temporally copied from simox control
        // Based on "A weighted least-norm solution based scheme for avoiding joint limits for redundant joint manipulators", T.F.Chan (1995)
        int nbJoints = jointAngles.size();
        Eigen::VectorXf weights(nbJoints);
        Eigen::VectorXf gradient = Eigen::VectorXf::Zero(nbJoints);

        if (jointAngleLimitGradient.rows() == 0)
        {
            jointAngleLimitGradient = Eigen::VectorXf::Zero(nbJoints);
        }

        for (int i = 0; i < nbJoints; i++)
        {
            if ((jointLimitsLow(i) == 0 && jointLimitsHigh(i) == 0) || jointLimitsHigh(i) > 1e8)
            {
                // Joint is limitless
                weights(i) = 1.;
            }
            else
            {
                gradient(i) = std::pow(jointLimitsHigh(i) - jointLimitsLow(i), 2) *
                              (2 * jointAngles(i) - jointLimitsHigh(i) - jointLimitsLow(i)) /
                              (4 * std::pow(jointLimitsHigh(i) - jointAngles(i), 2) *
                               std::pow(jointAngles(i) - jointLimitsLow(i), 2));

                if ((std::abs(gradient(i)) - std::abs(jointAngleLimitGradient(i))) >= 0.)
                {
                    // Joint going towards limits
                    weights(i) = 1. / std::sqrt((1. + std::abs(gradient(i))));
                }
                else
                {
                    // Joint going towards the center
                    weights(i) = 1.;
                }
            }
        }
        jointAngleLimitGradient = gradient;
        return weights.asDiagonal();
    }


} // namespace armarx::control::common::control_law
