/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/IK/DifferentialIK.h>

#include <RobotAPI/components/units/RobotUnit/RobotUnit.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValueForceTorque.h>

namespace armarx::control::common::ft
{
    /// TODO: this has to be moved to the robot model (e.g. the corresponding hand model)
    /// The values below corresponds to the Armar-6 hands,
    /// keep it here for now to continue working on controllers
    struct FTCalibrateDefaultValue
    {
        float tcpMassRight = 1.16064;
        Eigen::Vector3f tcpCoMInFTFrameRight {-0.00258287, -0.00108265, 0.09719};
        float tcpMassLeft = 1.14715;
        Eigen::Vector3f tcpCoMInFTFrameLeft {-0.00330519, 0.000470962, 0.100648};
    };


    class FTSensor
    {
    public:
        struct FTConfig
        {
            std::string sensorName;
            float ftFilter = 1.0f;
            float deadZoneForce = 0.0f;
            float deadZoneTorque = 0.0f;
            float timeLimit = 0.1f;

            //            float tcpMass = 0.0f;
            //            Eigen::Vector3f tcpCoMInFTSensorFrame;

            bool enableTCPGravityCompensation = false;
            bool reCalibrate = false;

            Eigen::Vector3f forceBaseline;
            Eigen::Vector3f torqueBaseline;
        };

        FTSensor(){};
        ~FTSensor(){};
        void initialize(
                const VirtualRobot::RobotNodeSetPtr& rns,
                const armarx::RobotUnitPtr& robotUnit,
                const FTConfig& c
                );
        void reset();
        bool calibrate(const FTConfig& c, double deltaT);
        Eigen::Vector6f& compensateTCPGravity(const FTConfig &data);
        Eigen::Vector6f& getFilteredForceTorque(const FTConfig &data);

        std::atomic_bool calibrated = false;

    private:
        std::string sensorName;
        std::string sensorFrame;

        FTCalibrateDefaultValue calibDefalts;

        const SensorValueForceTorque* forceSensor;
        VirtualRobot::RobotNodePtr sensorNode;

        Eigen::Vector6f tcpGravityCompensation;
        Eigen::Vector6f ft;

        /// local variables
        float tcpMass = 0.0f;
        Eigen::Vector3f tcpCoMInFTSensorFrame;

        /// for filtering and calibration
        Eigen::Vector3f forceOffset;
        Eigen::Vector3f torqueOffset;
        Eigen::Vector3f filteredForce;
        Eigen::Vector3f filteredTorque;
        Eigen::Vector3f filteredForceInRoot;
        Eigen::Vector3f filteredTorqueInRoot;
        float timeForCalibration;


        /// for gravity compensation
        const Eigen::Vector3f gravity { 0.0, 0.0, -9.8 };

        Eigen::Matrix4f forceFrameInTCP = Eigen::Matrix4f::Identity();
        Eigen::Vector3f tcpCoMInTCPFrame;

    };
}
