/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    TaskSpaceActiveImpedanceControl::ArmarXObjects::NJointTaskSpaceImpedanceController
 * @author     zhou ( you dot zhou at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "KeypointsImpedanceController.h"


namespace armarx::control::njoint_mp_controller::task_space
{
NJointControllerRegistration<NJointKeypointsImpedanceMPController> registrationControllerNJointKeypointsImpedanceMPController("NJointKeypointsImpedanceMPController");

NJointKeypointsImpedanceMPController::NJointKeypointsImpedanceMPController(
        const RobotUnitPtr& robotUnit,
        const NJointControllerConfigPtr& config,
        const VirtualRobot::RobotPtr& robot):
    NJointKeypointsImpedanceController{robotUnit, config, robot},
    mp::MPPool{}
{
    createMPs(configFileName);
}

std::string NJointKeypointsImpedanceMPController::getClassName(const Ice::Current &) const
{
    return "NJointKeypointsImpedanceMPController";
}

void NJointKeypointsImpedanceMPController::reconfigureController(const std::string &filename, const Ice::Current &)
{
    NJointKeypointsImpedanceController::reconfigureController(filename);
    reconfigureMPs(filename);
}


void NJointKeypointsImpedanceMPController::additionalTask()
{
    /// parse arguments and run mp
    for (auto mp: mps)
    {
        if (mp.second.mp->getClassName() == "KeypointsMP")
        {
            mp::KeypointsMPInputPtr in = std::dynamic_pointer_cast<mp::KeypointsMPInput>(mp.second.input);
            in->keypointPosition = controlStatusBuffer.getReadBuffer().filteredKeypointPosition;
            in->keypointVelocity = controlStatusBuffer.getReadBuffer().currentKeypointVelocity;
            in->deltaT           = controlStatusBuffer.getReadBuffer().deltaT;
            //            ARMARX_INFO << "additional Task: "  << VAROUT(in->keypointPosition) << VAROUT(in->keypointVelocity) << VAROUT(in->keypointVelocity.size());
        }
        else if (mp.second.mp->getClassName() == "JSMP")
        {
            mp::JSMPInputPtr in = std::dynamic_pointer_cast<mp::JSMPInput>(mp.second.input);
            in->angleRadian = controlStatusBuffer.getReadBuffer().qpos;
            in->angularVel  = controlStatusBuffer.getReadBuffer().qvel;
            in->deltaT      = controlStatusBuffer.getReadBuffer().deltaT;
        }
    }
    runMPs();
    for (auto mp: mps)
    {
        if (mp.second.mp->getClassName() == "KeypointsMP" and !mp.second.mp->isFinished())
        {
            mp::KeypointsMPOutputPtr out = std::dynamic_pointer_cast<mp::KeypointsMPOutput>(mp.second.output);
            getWriterControlStruct().desiredKeypointPosition = out->keypointPosition;
            getWriterControlStruct().desiredKeypointVelocity = out->keypointVelocity;
            writeControlStruct();
            //            ARMARX_INFO << "additional Task: "  << VAROUT(out->keypointPosition) << VAROUT(out->keypointVelocity) << VAROUT(out->keypointVelocity.size());
        }
        else if (mp.second.mp->getClassName() == "JSMP" and !mp.second.mp->isFinished())
        {
            mp::JSMPOutputPtr out = std::dynamic_pointer_cast<mp::JSMPOutput>(mp.second.output);
            getWriterControlStruct().desiredNullspaceJointAngles = out->angleRadian;
            writeControlStruct();
        }
    }
}

//void NJointKeypointsImpedanceMPController::deactivateController(const Ice::Current &)
//{
//    for (auto mp: mps)
//    {
//        mp.second.mp->stop();
//    }
//}

void NJointKeypointsImpedanceMPController::rtPreActivateController()
{
    ARMARX_IMPORTANT << "rt pre activate: reinitialize the mp input output, as well as the rt related buffer values";
    NJointKeypointsImpedanceController::rtPreActivateController();
    for (auto mp: mps)
    {
        if (mp.second.mp->getClassName() == "KeypointsMP")
        {
            ARMARX_IMPORTANT << "initialize input output buffers between KeypointsMP and keypoint controllers";

            auto in = std::dynamic_pointer_cast<mp::KeypointsMPInput>(mp.second.input);
            in->keypointPosition = controller.s.filteredKeypointPosition;
            in->keypointVelocity = controller.s.currentKeypointVelocity;

            auto out = std::dynamic_pointer_cast<mp::KeypointsMPOutput>(mp.second.output);
            out->keypointPosition = controller.s.filteredKeypointPosition;
            out->keypointVelocity = controller.s.currentKeypointVelocity;
            //            ARMARX_INFO << VAROUT(in->keypointPosition) << VAROUT(in->keypointVelocity) << VAROUT(out->keypointPosition) << VAROUT(out->keypointVelocity) << VAROUT(controller.s.numPoints);
        }
        else if (mp.second.mp->getClassName() == "JSMP")
        {
            ARMARX_IMPORTANT << "initialize input output buffers between JSMP and keypoint controllers";
            auto in  = std::dynamic_pointer_cast<mp::JSMPInput>(mp.second.input);
            auto out = std::dynamic_pointer_cast<mp::JSMPOutput>(mp.second.output);
            in->angleRadian  = controller.s.qpos;
            in->angularVel   = controller.s.qvel;
            out->angleRadian = controller.s.qpos;
            out->angularVel  = controller.s.qvel;
        }

    }
}
} /// namespace armarx::control::njoint_mp_controller::task_space
