/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    control::ArmarXObjects::retrieve_hand
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RetrieveHand.h"
#include <ArmarXCore/core/time/Metronome.h>
#include <armarx/control/deprecated_njoint_mp_controller/task_space/ControllerInterface.h>
#include <armarx/control/deprecated_njoint_mp_controller/task_space/NJointTSDMPController.h>
#include <armarx/control/common/utils.h>

namespace armarx::control::retrieve_hand::core
{
    using namespace armarx::control::deprecated_njoint_mp_controller::task_space;

    RetrieveHand::RetrieveHand(
        const armarx::control::retrieve_hand::core::RetrieveHand::Remote& remote,
        const armarx::control::retrieve_hand::core::RetrieveHand::Properties& properties) :
        remote(remote), properties(properties)
    {
    }

    bool
    RetrieveHand::_runTSVMPController()
    {
        ARMARX_IMPORTANT << "Creating TSVMP Controller for Retrieve hand skill";
        double phaseL = 100;
        double phaseK = 1000;
        double phaseDist0 = 50;
        double phaseDist1 = 10;
        double phaseKp = 2;
        double phaseKd = 1;
        double posToOriRatio = 10;
        double amp = 1;
        double maxLinearVel = 1000;
        double maxAngularVel = 10;


        armarx::NJointTaskSpaceDMPControllerConfigPtr tsConfig =
            new armarx::NJointTaskSpaceDMPControllerConfig(50,
                                                           20,
                                                           1,
                                                           "Linear",
                                                           "Discrete",
                                                           amp,
                                                           phaseL,
                                                           phaseK,
                                                           phaseDist0,
                                                           phaseDist1,
                                                           phaseKp,
                                                           phaseKd,
                                                           properties.timeDurationInSec,
                                                           posToOriRatio,
                                                           properties.robotNodeSet,
                                                           "",
                                                           "",
                                                           NJointTaskSpaceDMPControllerMode::eAll,
                                                           maxLinearVel,
                                                           maxAngularVel,
                                                           50.0,
                                                           "DMP",
                                                           properties.kpLinear,
                                                           properties.kdLinear,
                                                           properties.kpAngular,
                                                           properties.kdAngular,
                                                           1.0,
                                                           1.0);

        armarx::NJointTaskSpaceDMPControllerInterfacePrx dmpController =
            armarx::NJointTaskSpaceDMPControllerInterfacePrx::checkedCast(
                remote.robotUnit->createNJointController(
                    "NJointTSDMPController", "dmpController_RetrieveHand", tsConfig));


        std::vector<std::string> fileNames = properties.fileNames;
        dmpController->learnDMPFromFiles(fileNames);

        // load robot
        armarx::armem::robot_state::VirtualRobotReader robotReader;
        robotReader.connect(remote.memoryNameSystem);

        auto now = armarx::core::time::DateTime::Now();

        std::vector<double> goals;
        //        Eigen::Matrix4f currentPose = r->getRobotNodeSet(properties.robotNodeSet)->getTCP()->getPoseInRootFrame();
        goals = armarx::control::common::mat4ToDVec(properties.targetPose);
        ARMARX_INFO << "Running VMP with goal " << VAROUT(properties.targetPose) << ", vec: " << armarx::control::common::dVecToString(goals);

        // if (in.isViaPoseListSet() && in.isViaPoseCanValSet() && in.getViaPoseList().size() == in.getViaPoseCanVal().size())
        // {
        //     for (size_t i = 0; i < in.getViaPoseList().size(); ++i)
        //     {
        //         dmpController->setViaPoints(in.getViaPoseCanVal().at(i), Helpers::pose2dvec(in.getViaPoseList().at(i)->toEigen()));
        //     }
        // }

        dmpController->activateController();

        dmpController->runDMP(goals, 1.0);

        const auto metronomeTargetPeriod = armarx::Duration::MilliSeconds(10);
        armarx::Metronome loopScheduler(metronomeTargetPeriod);
        ARMARX_INFO << "VMP controller started";
        while (running.load() && not dmpController->isFinished())
        {
            // ARMARX_INFO << VAROUT(dmpController->getCanVal());
            loopScheduler.waitForNextTick();
        }

        ARMARX_DEBUG << "VMP Controller finished, try to delete controller";
        dmpController->deactivateController();
        while (dmpController->isControllerActive())
        {
            loopScheduler.waitForNextTick();
        }
        dmpController->deleteController();

        running.store(false);
        ARMARX_INFO << "Retrieve hand skill finished";
        return true;
    }

} // namespace armarx::control::retrieve_hand::core
